# flake8: noqa
from skale.contracts.skale_contract import SkaleContract
from skale.contracts.base_contract import transaction_method
from skale.contracts.etherbase.etherbase import Etherbase

__all__ = ['Etherbase']
